/*-------------------------------------------------------------------------+
| Copyright 2017 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.partitionmapping;

import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.MappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.copy.Copyable;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.random.RandomExploration;

/**
 * Encoding that maps partition, which host tasks, to execution units. Typically, these are cores.
 * The encoding is essentially a two-dimensional table.
 * 
 * @author diewald
 */
public class PartitionMappingEncoding extends
		MappingEncoding<PartitionMappingEncoding, Partition, IExecutionUnitAdapter<?>, PartitionMappingEntry>
		implements Copyable<PartitionMappingEncoding> {

	/** Strategy that defines how the initial mapping is constructed. */
	public enum CREATE_STARTEGY {
			/** One Partition per physical execution unit. */
			MIN_SET,
			/** Random number of partitions per physical execution unit. */
			RANDOM_SET,
			/**
			 * Same number of partitions per physical execution units as tasks exist in the system.
			 */
			MAX_SET
	}

	/** Constructor. */
	public PartitionMappingEncoding(Collection<IExecutionUnitAdapter<?>> targetSet) {
		super(targetSet);
	}

	/** Constructor. */
	// TODO: The strategy can be replaced by lambdas for a cleaner design.
	public PartitionMappingEncoding(Collection<IExecutionUnitAdapter<?>> targetSet, int numTasks,
			CREATE_STARTEGY strategy) {
		this(targetSet);
		// TODO: Move the creation strategy to the creator.
		// TODO: This does not consider task replicas yet.
		if(strategy == CREATE_STARTEGY.MIN_SET) {
			for(IExecutionUnitAdapter<?> target : targetSet) {
				Partition part = new Partition();
				allocationMap.put(part, new PartitionMappingEntry(part, target));
			}
		} else if(strategy == CREATE_STARTEGY.RANDOM_SET) {
			for(IExecutionUnitAdapter<?> target : targetSet) {
				int randMaxPartitions =
						RandomExploration.getInstance().nextInt(numTasks / targetSet.size());
				// Ensure at least one partition per phys. execution unit.
				randMaxPartitions = (randMaxPartitions < 1) ? 1 : randMaxPartitions;
				for(int i = 0; i < randMaxPartitions; ++i) {
					Partition part = new Partition();
					allocationMap.put(part, new PartitionMappingEntry(part, target));
				}
			}
		} else if(strategy == CREATE_STARTEGY.MAX_SET) {
			int parititonsPerResource = numTasks / targetSet.size();
			for(IExecutionUnitAdapter<?> target : targetSet) {
				for(int i = 0; i < parititonsPerResource; ++i) {
					Partition part = new Partition();
					allocationMap.put(part, new PartitionMappingEntry(part, target));
				}
			}
		}
	}

	/** Prototype Constructor. */
	private PartitionMappingEncoding() {
		// Prototype for newInstance().
		super(null);
	}

	/** Creates a {@link Partition} that is allocated to the given {@link IResourceAdapter}. */
	public void createPartition(IExecutionUnitAdapter<?> resource) {
		Partition newPartition = new Partition();
		allocationMap.put(newPartition, new PartitionMappingEntry(newPartition, resource));
	}

	/** Removes a random {@link Partition} allocation to the given {@link IResourceAdapter}. */
	public boolean removeRandomPartition(IExecutionUnitAdapter<?> resource) {
		List<Partition> partitions = getRequesters(resource);
		if(partitions.size() == 1) {
			return false;
		}
		int randIdx = RandomExploration.getInstance().nextInt(partitions.size());
		Partition rmPartition = partitions.get(randIdx);
		removePartition(resource, rmPartition);
		return true;
	}

	/** Removes the given {@link Partition} allocation to the given {@link IResourceAdapter}. */
	public void removePartition(IResourceAdapter<?> resource, Partition partition) {
		allocationMap.remove(resource, partition);
	}

	/** {@inheritDoc} */
	@Override
	public int size() {
		return allocationMap.keySet().size();
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public PartitionMappingEncoding newInstance() {
		return new PartitionMappingEncoding();
	}

	/** {@inheritDoc} */
	@Override
	public PartitionMappingEncoding copy() {
		PartitionMappingEncoding clonedEncoding = new PartitionMappingEncoding(getResources());

		Map<PartitionMappingEntry, PartitionMappingEntry> copyAssocMap = new HashMap<>();
		clonedEncoding.allocationMap = copyAllocationMap(copyAssocMap);

		// TODO: Implement a generic version of the assoc Map copy and call it here.
		// TODO: Implement a generic clone version, that can be called from here.

		return clonedEncoding;
	}
}
