/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene;

import java.lang.reflect.Method;

import org.aopalliance.intercept.MethodInterceptor;
import org.aopalliance.intercept.MethodInvocation;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.module.CreatorModule;
import org.fortiss.af3.exploration.alg.dse.module.DecoderModule;
import org.opt4j.core.Genotype;
import org.opt4j.core.problem.Decoder;

import com.google.common.collect.ClassToInstanceMap;
import com.google.inject.AbstractModule;
import com.google.inject.Injector;
import com.google.inject.Module;

/**
 * This {@link MethodInterceptor} is intended to be used with {@link CreatorModule}s or
 * {@link DecoderModule}s where the same instance of a {@link Genotype} or {@link Phenotype} may be
 * demanded multiple times in the same phase.
 * multiple ones "produce" the same {@link Decoder}.
 * It intercepts the decode method call (defined by an annotation) and launches all sub-
 * {@link DecoderModule}s before executing its own decoding. Sub-{@link DecoderModule}s are not
 * allowed to interfere with each other.
 * 
 * @author diewald
 */
public abstract class ProvisionInterceptor<T> extends AbstractModule implements MethodInterceptor {

	/**
	 * Reference to the {@link Injector} used to execute the create/decoding. Required to find all
	 * {@link Module}s that produce the same type.
	 */
	protected Injector injector;

	/** Encoding type: Returned by the invoked methods. E.g., Genotype, or Phenotype. */
	protected Class<T> encodingType;

	/** The type of the cache used to collect the results of the invoked method. */
	protected Class<? extends EncodingCache<T>> cacheType;

	/** Sets the {@link Injector} used in the create or decoding phase. */
	public void setInterceptParameters(Injector parentInjector,
			Class<? extends InjectorProvider> injectorProvider, Class<T> encodingType,
			Class<? extends EncodingCache<T>> cacheType) {
		this.injector = parentInjector.getInstance(injectorProvider).getInjector();
		this.encodingType = encodingType;
		this.cacheType = cacheType;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Object invoke(MethodInvocation invocation) throws Throwable {
		Method invokedMethod = invocation.getMethod();
		// Type erasure allows this cast. We're checking immediately afterwards.
		Class<T> returnType = (Class<T>)invokedMethod.getReturnType();
		if(!encodingType.isAssignableFrom(returnType)) {
			throw new Exception(
					"Could not create/decode: The return type from an operation is not a " +
							encodingType.getClass() + " or the method interceptor was bound to an" +
							" invalid method.");
		}

		// Check if the requested phenotype was already decoded. If so, return
		// this one instead.
		ClassToInstanceMap<T> aggregateMap = injector.getInstance(cacheType);
		T existingEncoding = aggregateMap.getInstance(returnType);
		if(existingEncoding != null && !invokedMethod.isAnnotationPresent(UpdatesEncoding.class)) {
			return existingEncoding;
		}

		// Execute suboperations before the "primary operation to update the encoding before
		// actually using it.
		// This feature is non-composable: We do not support updates for update methods to avoid
		// endless recursion here.
		T encoding = null;
		if(!invokedMethod.isAnnotationPresent(UpdatesEncoding.class)) {
			encoding = executeSubOperations(invocation, returnType);
		}

		// The return type of a decode method must be a Phenotype. It is the same method whose
		// execution is prevented by the overrides check above (return type 'Phenotype').
		encoding = (T)invocation.proceed();
		aggregateMap.putInstance(returnType, encoding);

		return encoding;
	}

	/**
	 * To be implemented for interceptors that support updating of an encoding. These are launched
	 * before the actual method is invoked (preprocessing). If this feature is not supported, just
	 * return {@code null}.
	 * Update methods must be annotated with @UpdatesEncoding.
	 * 
	 * @param invocation
	 *            Standard decoding method.
	 * @param returnType
	 *            expected return type: return type of the "standard" decoding method.
	 * @return Updated encoding.
	 * @throws Exception
	 *             to be throw if return types do not match.
	 */
	protected abstract T executeSubOperations(MethodInvocation invocation, Class<T> returnType)
			throws Exception;
}
