/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.safetyfunction;

import java.util.Arrays;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.constraint.InternalReplicationConstraint;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.constraint.InternalSeparationConstraint;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.UpdatesEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.Decodes;
import org.fortiss.af3.exploration.alg.dse.module.DecoderModule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISafetyFunctionAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.platform.hierarchic.model.tile.Tile;
import org.opt4j.core.Genotype;

import com.google.inject.Provides;

/**
 * Generates allocation and replication constraints for the safety architectures of the safety
 * function of a given {@link SafetyFunctionArchEncoding}.
 * 
 * @author diewald
 */
public class SFMappingConstraintDecoder<C> extends DecoderModule<AbstractTaskMappingEncoding> {

	/**
	 * Updates the input {@link AbstractTaskMappingEncoding} with additionally required allocation
	 * entries from the given safety architectures: Additional/Removed elements, replicated
	 * channels.
	 */
	@Provides
	@Decodes
	@UpdatesEncoding
	public AbstractTaskMappingEncoding updateComposable(SafetyFunctionArchEncoding sfaEnc,
			AbstractTaskMappingEncoding atmEnc) {
		for(SafetyFunctionArchEntry sfaEntry : sfaEnc.getAllEntries()) {
			addReplicationConstraints(sfaEntry, sfaEnc);
			addSeparationConstraints(sfaEntry, sfaEnc);
		}

		return atmEnc;
	}

	/**
	 * Add constraints on the number of task replia such that the number of {@link ITaskAdapter}s
	 * corresponds to the number of safety channels.
	 */
	private void addReplicationConstraints(SafetyFunctionArchEntry sFAEntry,
			SafetyFunctionArchEncoding sfaEnc) {
		int channelNumber = sFAEntry.getNumChannels();

		for(ITaskAdapter<?> comp : sFAEntry.getSafetyFunctionAdapter().getAllComponents()) {
			sfaEnc.putConstraint(InternalReplicationConstraint.class,
					new InternalReplicationConstraint(comp, channelNumber, channelNumber));
		}
	}

	/**
	 * Creates {@link InternalSeparationConstraint}s for the sets of realizing, diagnostic, and
	 * output units of each safety function.
	 */
	private void addSeparationConstraints(SafetyFunctionArchEntry sfaEntry,
			SafetyFunctionArchEncoding sfaEnc) {
		ISafetyFunctionAdapter<?> sfAdp = sfaEntry.getSafetyFunctionAdapter();

		// TODO: We should not be aware of the Tile type here.
		// FIXME: We need a "better" machanism than the fault containment region. Also,
		sfaEnc.putConstraint(InternalSeparationConstraint.class,
				new InternalSeparationConstraint(sfAdp.getRealizingComponents(), Tile.class));
		sfaEnc.putConstraint(InternalSeparationConstraint.class,
				new InternalSeparationConstraint(sfAdp.getDiagnosticComponents(), Tile.class));
		sfaEnc.putConstraint(InternalSeparationConstraint.class, new InternalSeparationConstraint(
				Arrays.asList(sfAdp.getOutputComponent()), Tile.class));
	}

	/** {@inheritDoc} */
	@Override
	public AbstractTaskMappingEncoding decode(Genotype genotype) {
		// TODO Auto-generated method stub
		return null;
	}
}
