/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.evaluator.constraint;

import static java.lang.Math.pow;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.stream.Collectors;

import org.eclipse.emf.common.util.Enumerator;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.FailSilentTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.evaluate.EvaluatorInjectorProvider;
import org.fortiss.af3.exploration.alg.dse.module.EvaluatorModule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.moea.model.predefined.SafetyIntegrityLevelConstraint;
import org.fortiss.af3.safety.model.SIL;
import org.opt4j.core.DoubleValue;
import org.opt4j.core.Value;
import org.opt4j.core.problem.Evaluator;

import com.google.common.collect.ClassToInstanceMap;
import com.google.inject.Provides;
import com.google.inject.name.Named;

/**
 * Preliminary {@link Evaluator} to check whether the allocation of specific {@link ITaskAdapter}s
 * to {@link IExecutionUnitAdapter}s satisfy the Safety
 * Integrity Level required by the {@link ITaskAdapter}.
 * 
 * @author diewald
 */
public class SafetyIntegrityLevelConstraintEvaluator<S extends InstantiatedTaskMappingEntry, T extends InstantiatedTaskMappingEncoding<S>>
		extends EvaluatorModule<SIL, SafetyIntegrityLevelConstraint, T>
		implements IConstraintEvaluator<SIL, SafetyIntegrityLevelConstraint, T> {

	/** Constructor. */
	public SafetyIntegrityLevelConstraintEvaluator(
			Class<SafetyIntegrityLevelConstraint> explorationTargetType) throws Exception {
		super(explorationTargetType);
	}

	/** Constructor. */
	public SafetyIntegrityLevelConstraintEvaluator(SafetyIntegrityLevelConstraint explorationTarget)
			throws Exception {
		super(explorationTarget);
	}

	/** {@inheritDoc} */
	@Provides
	@Named("SafetyIntegrityLevelConstraintEvaluator")
	public DoubleValue evaluateGoal(FailSilentTaskMappingEncoding phenotype,
			ClassToInstanceMap<EObject> transformedModels) {
		// FIXME: get Rid of the cast, once the genric/interface problem is solved in guice.
		List<SIL> silViolations = getSILViolations((T)phenotype);
		double aggregatedViolations =
				silViolations.stream().mapToDouble(sil -> pow(sil.getValue(), 2.0)).sum();

		return new DoubleValue(aggregatedViolations);
	}

	/** {@inheritDoc} */
	@Override
	public boolean validate(ExplorationConstraint<?> expTarget,
			EvaluatorInjectorProvider evalInjProvider, Value<?> value) {
		FailSilentTaskMappingEncoding fstmEnc =
				evalInjProvider.getInjector().getInstance(FailSilentTaskMappingEncoding.class);
		List<ITaskAdapter<?>> silComponents = getSILComponentAdapters((T)fstmEnc);
		List<SIL> silBounds = explorationTarget.getLimitation();
		for(int idx = 0; idx < silComponents.size(); idx++) {
			ITaskAdapter<?> silComponent = silComponents.get(idx);
			SIL actualSIL = detectSIL(silComponent,
					(Collection<S>)fstmEnc.getMappingEntriesOf(silComponent));
			if(actualSIL.getValue() < silBounds.get(idx).getValue()) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Returns the list of of violated {@link SIL} requirements as the difference between the
	 * required {@link SIL} and the actually achieved {@link SIL}.
	 * The list has the same order as the list of referenced {@link Component}s in the
	 * {@link SafetyIntegrityLevelConstraint}.
	 */
	private List<SIL> getSILViolations(T mapping) {
		List<SIL> silViolations = new ArrayList<>();
		List<ITaskAdapter<?>> silComponents = getSILComponentAdapters(mapping);
		List<SIL> silBounds = explorationTarget.getLimitation();
		for(int idx = 0; idx < silComponents.size(); idx++) {
			ITaskAdapter<?> silComponent = silComponents.get(idx);
			SIL actualSIL = detectSIL(silComponent, mapping.getMappingEntriesOf(silComponent));
			silViolations.add(calculateSILViolation(silBounds.get(idx), actualSIL));
		}
		return silViolations;
	}

	/**
	 * Returns the set of {@link ITaskAdapter}s that is equivalent to the set of {@link Component}s
	 * given in the {@link SafetyIntegrityLevelConstraint}.
	 */
	private List<ITaskAdapter<?>> getSILComponentAdapters(T mapping) {
		Collection<ITaskAdapter<?>> tasks = mapping.getRequesters();
		return tasks.stream().filter(c -> explorationTarget.getComponents().contains(c.getObject()))
				.collect(Collectors.toList());
	}

	// TODO: Port to org.fortiss.af3.safety
	/** Helper function for the calculation of the actual SIL of a deployed component. */
	private SIL detectSIL(@SuppressWarnings("unused") ITaskAdapter<?> deployableComponent,
			Collection<S> allEntries) {
		// Determine the number of replications onto different execution units. This separation is
		// yet very primitive.
		Collection<IExecutionUnitAdapter<?>> hwReplica = new HashSet<IExecutionUnitAdapter<?>>();
		for(S currentTaskReplic : allEntries) {
			hwReplica.add(currentTaskReplic.getTarget());
		}
		int silTypeToInt = 0;

		// Determine the SIL of a "safety function", i.e. a safety-annotated component mapped onto
		// an execution unit, according to the tables given in IEC 61508-2.
		List<SIL> parallelSILs = new ArrayList<SIL>();
		// for(IExecutionUnitAdapter<?> ecu : hwReplica) {
		// TODO: Update the SafetyAnalysis Evaluator with comparable SILs.
		// double sff = ecu.getSafeFailureFraction();
		// if(sff >= 0.0 && sff < 0.6) {
		// parallelSILs.add(SIL.SIL0);
		// } else if(sff < 0.9) {
		// parallelSILs.add(SIL.SIL1);
		// } else if(sff < 0.99) {
		// parallelSILs.add(SIL.SIL2);
		// } else if(sff <= 1.0) {
		// parallelSILs.add(SIL.SIL3);
		// } else {
		// throw new RuntimeException(
		// "An invalid value for the safe failure fraction has been given to the optimization.
		// The allowed range is 0 to 1.");
		// }
		// }

		// Determine the overall SIL of the safety function considering the hardware replication.
		// The upper limit is set to 4, in accordance with IEC 61508.
		for(SIL sil : parallelSILs) {
			silTypeToInt += sil.getValue();
		}
		if(silTypeToInt > 4)
			silTypeToInt = 4;

		return SIL.get(silTypeToInt);
	}

	/**
	 * Helper function for the calculation of the punishment value for a single SIL violation.
	 * Currently, this value is determined by the iteration number (weak punishment in the
	 * beginning) and the quadratical deviation from the desired SIL.
	 */
	// TODO: Port to org.fortiss.af3.safety
	private SIL calculateSILViolation(Enumerator requiredSIL, SIL actualSIL) {
		int differenceSIL = requiredSIL.getValue() - actualSIL.getValue();
		return SIL.get(differenceSIL);

		// TODO divide by number of iterations, currently not accessible
		// if(differenceSIL > 0) {
		// return getIterationCountAndIncrement() * Math.pow(differenceSIL, 2.0);
		// }
		// return 0.0;
	}
}
