/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.evaluator.objective;

import java.util.Collection;

import org.apfloat.Apfloat;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.FailSilentTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.FailSilentTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.TaskGraph;
import org.fortiss.af3.exploration.moea.model.predefined.FailureMinObjective;
import org.fortiss.af3.exploration.moea.model.predefined.FailureType;
import org.opt4j.core.DoubleValue;

import com.google.inject.Provides;
import com.google.inject.name.Named;

/**
 * Reliability analysis for strict time-triggered execution model (with fail-silent assumption).
 * 
 * @author huang
 */
public class FailSilentReliabilityEvaluator extends
		ReliabilityEvaluatorBase<FailSilentTaskMappingEntry, FailSilentTaskMappingEncoding> {

	/** Creates a fail-silent reliability analysis evaluator for a given exploration goal. */
	public FailSilentReliabilityEvaluator(FailureMinObjective goal) throws Exception {
		super(goal);
	}

	/**
	 * Evaluates the reliability of an {@link ITaskAdapter} given the current mapping and schedule.
	 */
	@Provides
	@Named("FailSilentReliabilityEvaluator")
	public DoubleValue evaluate(
			StrictTTSchedule<FailSilentTaskMappingEntry, FailSilentTaskMappingEncoding> schedule,
			FailSilentTaskMappingEncoding itmEnc, TaskGraph taskGraph) {
		FailureType failureProbabilityType = explorationTarget.getType();

		// for the StrictTTFaileSilence model, no sdc could occur
		if(failureProbabilityType == FailureType.SDC) {
			return new DoubleValue(0D);
		}

		Apfloat overallSuccRate = new Apfloat(1, FP_PRECISION);
		Apfloat one = new Apfloat(1, FP_PRECISION);

		for(ITaskAdapter<?> deployableComponent : itmEnc.getRequesters()) {
			// TODO: Check if this correctly determines the taskFailureRate

			// only software tasks are considered

			// the failure rate of the single request
			Apfloat taskFailureRate = new Apfloat(1, FP_PRECISION);
			// get the set of mappings
			Collection<FailSilentTaskMappingEntry> entries =
					itmEnc.getMappingEntriesOf(deployableComponent);
			for(FailSilentTaskMappingEntry entry : entries) {
				Apfloat fr = getComponentFailureRate(deployableComponent, entry.getTarget());
				if(fr != null) {
					// the task fails only if all of its instances fail
					// ASSUMPTIN: fail-silence
					taskFailureRate = taskFailureRate.multiply(fr);
				}
			}
			// the overall application succeeds only if all of its tasks succeed
			Apfloat taskSuccRate = one.subtract(taskFailureRate);
			overallSuccRate = overallSuccRate.multiply(taskSuccRate);
		}

		// return value is the overall failure probability
		return new DoubleValue(one.subtract(overallSuccRate).doubleValue());
	}
}
