/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.service;

import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map.Entry;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;

/**
 * Allows to capture the already created/decoded {@link Phenotype}s and pass them to their successor
 * creators/decoders as a flat map that associates a {@link Phenotype}s type with its instance.
 * Using a {@link Phenotype} for this job allows to "protect" the internal map from
 * unintended read/write operations.
 * 
 * @author diewald
 */
public class ExplorationEncodingMap<T extends IExplorationEncoding>
		implements Iterable<Entry<Class<? extends T>, T>> {
	/**
	 * Map which correlates the {@link IExplorationEncoding} classes with their corresponding
	 * {@link IExplorationEncoding}s.
	 */
	protected LinkedHashMap<Class<? extends T>, T> encodings = new LinkedHashMap<>();

	/**
	 * Returns whether {@code this} {@link ExplorationEncodingMap} contains an encoding that is
	 * assignable from the given encoding type.
	 */
	public boolean containsEncoding(Class<? extends T> encodingType) {
		if(getEncodingInternal(encodingType) != null) {
			return true;
		}
		return false;
	}

	/**
	 * Returns the {@link IExplorationEncoding} of the given type. If no
	 * {@link IExplorationEncoding} of the given type
	 * can be found, {@code null} is returned.
	 */
	public <E extends T> E getEncoding(Class<?> subGenotypeClass) {
		// FIXME: Temp HACK to simplify the decoder dependency management. This map will be removed
		// during the rework.
		return getEncodingInternal((Class<E>)subGenotypeClass);
	}

	/**
	 * Internal helper method to retrieve a sub-{@link IExplorationEncoding} by its type hierarchy.
	 */
	protected <E extends IExplorationEncoding> E getEncodingInternal(Class<E> encodingType) {
		for(Entry<Class<? extends T>, T> composablePheno : encodings.entrySet()) {
			if(encodingType.isAssignableFrom(composablePheno.getKey())) {
				@SuppressWarnings("unchecked") E retPhenotype = (E)composablePheno.getValue();
				return retPhenotype;
			}
		}
		return null;
	}

	/**
	 * Adds the {@link IExplorationEncoding} to {@code this} map and uses the given type as its
	 * identifier.
	 */
	public void putEncoding(Class<? extends T> subSeqClass, T subSeqPheno) {
		encodings.put(subSeqClass, subSeqPheno);
	}

	/** {@inheritDoc} */
	@Override
	public Iterator<Entry<Class<? extends T>, T>> iterator() {
		return encodings.entrySet().iterator();
	}

	/**
	 * Returns the set of all identifiers that are related with the {@link IExplorationEncoding}s in
	 * {@code this} map.
	 */
	public Set<Class<? extends T>> keySet() {
		return encodings.keySet();
	}

	/**
	 * Adds all entries of the given {@link ExplorationEncodingMap} to {@code this}
	 * {@link ExplorationEncodingMap}.
	 */
	public void addAll(ExplorationEncodingMap<T> addMap) {
		addMap.forEach(e -> encodings.put(e.getKey(), e.getValue()));
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		String ownName = getClass().getSimpleName() + ": \n";
		StringBuilder strBuilder = new StringBuilder(ownName);
		encodings.keySet().forEach(p -> strBuilder.append("-> " + p.getSimpleName() + "\n"));
		return strBuilder.toString();
	}
}
