/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.util;

import static org.fortiss.af3.exploration.util.ExplorationDslModelElementFactory.createObjectiveExpression;
import static org.fortiss.af3.exploration.util.ExplorationDslModelElementFactory.createSummationOp;
import static org.fortiss.tooling.kernel.utils.UniqueIDUtils.prepareUniqueID;

import org.eclipse.emf.common.util.EList;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.exploration.dsl.model.expressions.Expression;
import org.fortiss.af3.exploration.dsl.model.expressions.ObjectiveExpression;
import org.fortiss.af3.exploration.dsl.model.operators.LocationOp;
import org.fortiss.af3.exploration.dsl.model.operators.LocationOperatorEnum;
import org.fortiss.af3.exploration.dsl.model.operators.OperatorsFactory;
import org.fortiss.af3.exploration.dsl.model.operators.OptimizationDirectionEnum;
import org.fortiss.af3.exploration.dsl.model.operators.SummationOp;
import org.fortiss.af3.exploration.dsl.model.patterns.LocationProperty;
import org.fortiss.af3.exploration.dsl.model.patterns.PatternsFactory;
import org.fortiss.af3.exploration.model.ExplorationFactory;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.fortiss.af3.exploration.moea.model.MoeaFactory;
import org.fortiss.af3.exploration.moea.model.parameters.EAOptimizer;
import org.fortiss.af3.exploration.moea.model.parameters.ExecutionModel;
import org.fortiss.af3.exploration.moea.model.parameters.ParametersFactory;
import org.fortiss.af3.exploration.moea.model.parameters.SystemParameters;
import org.fortiss.af3.exploration.moea.model.predefined.ComponentMultiAllocationConstraint;
import org.fortiss.af3.exploration.moea.model.predefined.ComponentMultiDislocationConstraint;
import org.fortiss.af3.exploration.moea.model.predefined.EnergyMinObjective;
import org.fortiss.af3.exploration.moea.model.predefined.FailureMinObjective;
import org.fortiss.af3.exploration.moea.model.predefined.FailureType;
import org.fortiss.af3.exploration.moea.model.predefined.PredefinedFactory;
import org.fortiss.af3.exploration.moea.model.predefined.ReliabilityAnalysis;
import org.fortiss.af3.exploration.moea.model.predefined.SafetyIntegrityLevelConstraint;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.service.IPersistencyService;

/**
 * Model element factory for design space exploration.
 * 
 * @author diewald
 */
public class DesignSpaceExplorationModelElementFactory {

	/** Creates an instance of a {@link DseSpecification} with default parameters. */
	public static DseSpecification createDseSpecification(String name) {
		DseSpecification dse = MoeaFactory.eINSTANCE.createDseSpecification();
		dse.setName(name);
		return dse;
	}

	/** Creates a DseSpecification referencing the given {@link ExplorationSpecification}. */
	public static DseSpecification
			createDefaultDesignSpaceExploration(ExplorationSpecification expSpec) {
		ITopLevelElement modelContext =
				IPersistencyService.getInstance().getTopLevelElementFor(expSpec);
		return createDefaultDesignSpaceExploration(expSpec, modelContext);
	}

	/**
	 * Creates a DseSpecification and adds the {@link ExplorationSpecification} to it using the
	 * given {@code modelContext}.
	 */
	public static DseSpecification createDefaultDesignSpaceExploration(
			ExplorationSpecification expSpec, ITopLevelElement modelContext) {
		DseSpecification dseSpec = createDseSpecification("Default DSE model",
				ExecutionModel.TT_FAIL_SILENCE, createDefaultEAOptimizer());
		modelContext.runAsCommand(() -> dseSpec.setTargetSpecification(expSpec));
		return dseSpec;
	}

	/** Creates an instance of a {@link DseSpecification} with default parameters. */
	public static DseSpecification createDefaultDesignSpaceExploration() {
		return createDseSpecification("Default DSE model", ExecutionModel.TT_FAIL_SILENCE,
				createDefaultEAOptimizer());
	}

	/** Creates an instance of a {@link DseSpecification} with the given parameters. */
	public static DseSpecification createDseSpecification(String name, ExecutionModel execModel,
			EAOptimizer eaOptimizer) {
		DseSpecification dse = createDseSpecification(name);
		dse.setParameters(createSystemParameters(execModel));
		dse.setOptimizer(eaOptimizer);
		// A goal specification does not require any default values, so created here.
		dse.setTargetSpecification(ExplorationFactory.eINSTANCE.createExplorationSpecification());
		return dse;
	}

	/** "Wrapper" for creating an instance of default DSE parameters for the model. */
	public static SystemParameters createSystemParameters(ExecutionModel execModel) {
		SystemParameters dseParams = ParametersFactory.eINSTANCE.createSystemParameters();
		dseParams.setExecutionModel(execModel);
		return dseParams;
	}

	/** Creates an default EA optimizer configuration */
	public static EAOptimizer createDefaultEAOptimizer() {
		// Parameter defaults are preset in exploration.ecore.
		EAOptimizer optimizer = ParametersFactory.eINSTANCE.createEAOptimizer();
		return optimizer;
	}

	/**
	 * Creates an energy minimization objective within the goal specification of the current DSE
	 * model.
	 */
	public static EnergyMinObjective addEnergyMinObjective(ExplorationSpecification targetSpec) {
		EnergyMinObjective energyObjective = createEnergyMinObjective();
		ITopLevelElement modelContext =
				IPersistencyService.getInstance().getTopLevelElementFor(targetSpec);
		prepareUniqueID(energyObjective, targetSpec);
		modelContext.runAsCommand(() -> targetSpec.getTargets().add(energyObjective));
		return energyObjective;
	}

	/**
	 * Creates an {@link EnergyMinObjective} that minimizes the sum of the energy consumption that
	 * is specified by {@link LocationProperty}s.
	 */
	public static EnergyMinObjective createEnergyMinObjective() {
		EnergyMinObjective energyObjective = PredefinedFactory.eINSTANCE.createEnergyMinObjective();
		energyObjective.setName("Energy Minimization");
		energyObjective.setPriority(0);
		ObjectiveExpression optExpr = createObjectiveExpression(OptimizationDirectionEnum.MIN);
		optExpr.setRight(createEnergyConsumptionSumExpression());
		energyObjective.setMoeaExpression(optExpr);
		return energyObjective;
	}

	/**
	 * Creates an {@link Expression} that provides the summed energy consumption of all allocated
	 * components.
	 */
	public static SummationOp<Double> createEnergyConsumptionSumExpression() {
		SummationOp<Double> sumOp = createSummationOp();
		LocationProperty<Component, ExecutionUnit, Double> allocProp =
				PatternsFactory.eINSTANCE.createLocationProperty();
		LocationOp<Component, ExecutionUnit> allocOp =
				OperatorsFactory.eINSTANCE.createLocationOp();
		allocOp.setOperator(LocationOperatorEnum.ALLOCATION);
		allocProp.setLocOp(allocOp);
		sumOp.setRight(allocProp);
		return sumOp;
	}

	/**
	 * Creates a {@link FailureMinObjective} and adds it to the given
	 * {@link ExplorationSpecification}.
	 */
	public static FailureMinObjective addFailureMinObjective(ExplorationSpecification targetSpec) {
		FailureMinObjective failureMinObjective = createFailureMinObjective();
		ITopLevelElement modelContext =
				IPersistencyService.getInstance().getTopLevelElementFor(targetSpec);
		prepareUniqueID(failureMinObjective, targetSpec);
		modelContext.runAsCommand(() -> targetSpec.getTargets().add(failureMinObjective));
		return failureMinObjective;
	}

	/** Creates a default {@link FailureMinObjective}. */
	public static FailureMinObjective createFailureMinObjective() {
		FailureMinObjective failureMinObjective =
				PredefinedFactory.eINSTANCE.createFailureMinObjective();
		failureMinObjective.setName("Failure Minimization");
		failureMinObjective.setType(FailureType.ALL);
		failureMinObjective.setPriority(1);
		failureMinObjective.setMoeaExpression(createFailureMinExpression());
		return failureMinObjective;
	}

	/**
	 * Creates an {@link ObjectiveExpression} for {@link FailureMinObjective}s that aims to minimize
	 * the failure probability based on a {@link ReliabilityAnalysis}.
	 */
	public static ObjectiveExpression createFailureMinExpression() {
		ReliabilityAnalysis relAnalysis = createReliabilityAnalysis();
		ObjectiveExpression objectiveExpr =
				createObjectiveExpression(OptimizationDirectionEnum.MIN);
		objectiveExpr.setRight(relAnalysis);
		return objectiveExpr;
	}

	/**
	 * Creates an {@link ReliabilityAnalysis} to analyze the reliability of the given
	 * {@link Component}. Used in {@link FailureMinObjective}s .
	 */
	public static ReliabilityAnalysis createReliabilityAnalysis() {
		ReliabilityAnalysis relAnalysis = PredefinedFactory.eINSTANCE.createReliabilityAnalysis();
		return relAnalysis;
	}

	/**
	 * Creates a fixed deployment constraint within the goal specification of the current DSE model.
	 */
	public static ComponentMultiAllocationConstraint
			addMultiComponentAllocationConstraint(ExplorationSpecification targetSpec) {
		ComponentMultiAllocationConstraint fixedDeploymentConstraint =
				createComponentMultiAllocationConstraint();
		prepareUniqueID(fixedDeploymentConstraint, targetSpec);
		ITopLevelElement modelContext =
				IPersistencyService.getInstance().getTopLevelElementFor(targetSpec);
		modelContext.runAsCommand(() -> targetSpec.getTargets().add(fixedDeploymentConstraint));
		return fixedDeploymentConstraint;
	}

	/** Creates an {@link ComponentMultiAllocationConstraint} constraint. */
	public static ComponentMultiAllocationConstraint createComponentMultiAllocationConstraint(
			Component comp, EList<ExecutionUnit> execUnits) {
		ComponentMultiAllocationConstraint mCompAllocationConstraint =
				createComponentMultiAllocationConstraint();
		mCompAllocationConstraint.setComponentMultiAllocations(comp, execUnits);
		return mCompAllocationConstraint;
	}

	/** Creates an "empty" {@link ComponentMultiAllocationConstraint} constraint. */
	public static ComponentMultiAllocationConstraint createComponentMultiAllocationConstraint() {
		ComponentMultiAllocationConstraint mCompAllocationConstraint =
				PredefinedFactory.eINSTANCE.createComponentMultiAllocationConstraint();
		mCompAllocationConstraint.setName("Component Allocation Constraint (1:n)");
		return mCompAllocationConstraint;
	}

	/**
	 * Creates a {@link ComponentMultiDislocationConstraint} to the given
	 * {@link ExplorationSpecification}.
	 */
	public static ComponentMultiDislocationConstraint
			addComponentMultiDislocationConstraint(ExplorationSpecification targetSpec) {
		ComponentMultiDislocationConstraint multiComponentDislocationConstraint =
				createComponentMultiDislocationConstraint();
		prepareUniqueID(multiComponentDislocationConstraint, targetSpec);
		ITopLevelElement modelContext =
				IPersistencyService.getInstance().getTopLevelElementFor(targetSpec);
		modelContext.runAsCommand(
				() -> targetSpec.getTargets().add(multiComponentDislocationConstraint));
		return multiComponentDislocationConstraint;
	}

	/** Creates an "empty" {@link ComponentMultiDislocationConstraint}. */
	public static ComponentMultiDislocationConstraint createComponentMultiDislocationConstraint() {
		ComponentMultiDislocationConstraint multiComponentDislocationConstraint =
				PredefinedFactory.eINSTANCE.createComponentMultiDislocationConstraint();
		multiComponentDislocationConstraint.setName("Multi Component Dislocation Constraint (1:n)");
		return multiComponentDislocationConstraint;
	}

	/**
	 * Creates a safety integrity level constraint within the goal specification of the current DSE
	 * model.
	 */
	public static SafetyIntegrityLevelConstraint
			addSafetyIntegrityLevelConstraint(ExplorationSpecification targetSpec) {
		SafetyIntegrityLevelConstraint safetyIntegrityLevelConstraint =
				PredefinedFactory.eINSTANCE.createSafetyIntegrityLevelConstraint();
		safetyIntegrityLevelConstraint.setName("Safety Integrity Level Constraint");
		prepareUniqueID(safetyIntegrityLevelConstraint, targetSpec);
		ITopLevelElement modelContext =
				IPersistencyService.getInstance().getTopLevelElementFor(targetSpec);
		modelContext
				.runAsCommand(() -> targetSpec.getTargets().add(safetyIntegrityLevelConstraint));
		return safetyIntegrityLevelConstraint;
	}
}
